/*
 * ulimits - User limits utility
 * Copyright (C) 2012,2013,2015,2016 Damien Goutte-Gattat
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <err.h>
#include <getopt.h>

#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/resource.h>


/* Help and informations about the program. */

static void
usage(int status)
{
    puts("Usage: ulimits [options] [command...]\n\
Parse " LIMITS_FILE " and apply resource limits for the calling user.\n\
If a command is specified, it is executed with the applied limits.\n");

    puts("Options:\n\
  -h, --help            Display this help message.\n\
  -v, --version         Display the version message.\n\
");

    puts("\
  -c, --ulimit          Print an ulimit command instead of setting\n\
                        the limits directly.\n\
  -s, --set             Set the limits defined in limits file.\n\
  -p, --process PID     Set limits for the specified process instead\n\
                        of the calling process.\n\
");

    printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);

    exit(status);
}

static void
info(void)
{
    printf("\
ulimits %s\n\
Copyright (C) 2012,2013,2015,2016 Damien Goutte-Gattat\n\
\n\
This program is released under the GNU General Public License.\n\
See the COPYING file or <http://www.gnu.org/licenses/gpl.html>.\n\
", VERSION);

    exit(EXIT_SUCCESS);
}


/* Resource limits stuff. */

/**
 * Represents a limit specification.
 */
struct rlimit_setting {
    /** The resource identifier (setrlimit's first parameter). */
    int     resource;

    /** The hard limit for that resource. */
    rlim_t  limit;
};

/* Pseudo resource limit. */
#define PLIMIT_UMASK    RLIMIT_NLIMITS + 1
#define PLIMIT_LOGIN    RLIMIT_NLIMITS + 2
#define PLIMIT_PRIORITY RLIMIT_NLIMITS + 3

/**
 * Check if user is in the group.
 *
 * @param uname A pointer to a username.
 * @param gname A pointer to a groupname.
 * @return 0 if a user belongs to group, -1 otherwise.
 */
static int
user_in_group(const char *uname, const char *gname)
{
    struct group *groupdata;
    char *const *list;

    if ( uname == NULL || gname == NULL )
        return -1;

    if ( *gname == '@' )
        gname += 1;
    else
        return -1;

    if ( (groupdata = getgrnam(gname)) == NULL )
        return -1;

    list = groupdata->gr_mem;

    while ( *list != NULL ) {
        if ( strcmp(*list, uname) == 0 )
            return 0;

        list += 1;
    }

    return -1;
}

/**
 * Parses a single limit specification.
 * This function parses a single limit specification (that is, a single
 * letter followed by a numerical value) and fills the provided
 * rlimit_setting structure. The string pointer is then moved to the
 * position past the parsed specification.
 *
 * @param s A pointer to the string to parse.
 * @param rlimit A pointer to a struct rlimit_setting to fill.
 * @return 0 for a correctly parsed limit, or -1 if an error occured.
 */
static int
parse_limit_value(char **s, struct rlimit_setting *rlimit)
{
    char *c, *p;
    unsigned long value;

    if ( s == NULL || *s == NULL || rlimit == NULL ) {
        errno = EINVAL;
        return -1;
    }

    /* Skip leading whitespaces. */
    c = *s;
    while ( *c == ' ' )
        c += 1;

    /* Get the numerical value for the limit. */
    if ( *(c + 1) == '-' ) {
        value = RLIM_INFINITY;
        p = c + 2;
    }
    else {
        errno = 0;
        /* The value for K (umask) shall be read as a
         * number in base 8. */
        value = strtol(c + 1, &p, *c == 'K' ? 8 : 10);
        if ( errno || p == c + 1 )
            return -1;
    }

    /* Select limit to set. */
    switch ( *c ) {
#define ULIMIT(symbol, letter, unused1, factor, unused2)                    \
    case letter:                                                            \
        rlimit->resource = symbol;                                          \
        rlimit->limit = value == RLIM_INFINITY ? value : value * factor;    \
        break;
#include "ulimits.h"
#undef ULIMIT

    default:
        return -1;
    }

    /* Move the string pointer past the parsed limit. */
    *s = p;

    return 0;
}

/**
 * Parses a limit string in the limits(5) format.
 * Parsed limits are stored into the provided array, which is
 * terminated by a pseudo-limit identified by -1.
 *
 * @param limit_string The string to parse.
 * @param rlimits An array of struct rlimit_setting to fill.
 * @param len The size of the \a rlimits array.
 * @return The number of parsed limits, or -1 if an error occured.
 */
static ssize_t
parse_limit_string(const char              *limit_string,
                   struct rlimit_setting   *rlimits,
                   size_t                   len)
{
    char *s = (char *)limit_string;
    size_t n = 0;

    if ( limit_string == NULL || rlimits == NULL ) {
        errno = EINVAL;
        return -1;
    }

    /* Skip whitespaces. */
    while ( *s == ' ' || *s == '\t' )
        s += 1;

    /* A single dash means no limits. */
    if ( *s == '-' && *(s + 1) == '\0' ) {
        for ( ; n < RLIMIT_NLIMITS && n < len - 1 ; n++ ) {
            rlimits[n].resource = n;
            rlimits[n].limit = RLIM_INFINITY;
        }
        s += 1;
    }

    while ( *s != '\0' && n < len - 1 ) {
        if ( parse_limit_value(&s, &(rlimits[n])) < 0 )
            return -1;

        n += 1;
    }

    return n;
}

/**
 * Discards characters from the specified stream up to the next
 * end of line.
 *
 * @param f The stream to read from.
 * @return 0 if successful, or -1 if an error occured.
 */
static int
discard_line(FILE *f)
{
    int c;

    if ( f == NULL ) {
        errno = EINVAL;
        return -1;
    }

    while ( (c = fgetc(f)) != '\n' && c != EOF ) ;

    return c == EOF ? -1 : 0;
}

/**
 * Reads a single line from the specified stream.
 * This functions differs from standard fgets(3) in two aspects: the
 * newline character is not stored, and if the line to read is too long
 * to fit into the provided buffer, the whole line is discarded.
 *
 * @param f The stream to read from.
 * @param buffer A character buffer to be filled with the read line.
 * @param len Size of the \a buffer array.
 * @return The number of characters read, or -1 if an error occured.
 */
static ssize_t
get_line(FILE *f, char *buffer, size_t len)
{
    int c;
    size_t n = 0;

    if ( f == NULL || buffer == NULL ) {
        errno = EINVAL;
        return -1;
    }

#ifdef ENABLE_MERGED_LIMITS
    if ( (n = strlen(buffer)) > 0 )
        buffer[n++] = ' ';
#endif

    while ( (c = fgetc(f)) != '\n' ) {
        if ( c == EOF )
            return -1;

        if ( n >= len - 1 ) {   /* line too long */
            discard_line(f);
            return -1;
        }

        buffer[n++] = (char)c;
    }
    buffer[n] = '\0';

    return n;
}

/**
 * Reads from the specified file up to the first space.
 * If the field is too long to fit into the provided buffer, the whole
 * line is regarded as invalid and discarded, and reading continues
 * with the next line until a valid field is read.
 *
 * @param f The stream to read from.
 * @param buffer A character buffer to be filled with the read data.
 * @param len Size of the \a buffer array.
 * @return The number of characters read, or -1 if an error occured.
 */
static ssize_t
get_first_field(FILE *f, char *buffer, size_t len)
{
    int c, found = 0;
    size_t n = 0;

    if ( f == NULL || buffer == NULL ) {
        errno = EINVAL;
        return -1;
    }

    while ( ! found && ! feof(f) ) {
        c = fgetc(f);

        if (c == '#')
            discard_line(f);
        else if ( isgraph(c) && n < len - 1 )
            buffer[n++] = (char)c;
        else if ( (c == ' ' || c == '\t') && n > 0 )
            found = 1;
        else {
            n = 0;
            if ( c != '\n' )
                discard_line(f);
        }
    }
    buffer[n] = '\0';

    return n > 0 ? (ssize_t)n : -1;
}

static FILE *
open_limits_file(const char *filename)
{
    struct stat buf;
    FILE *f = NULL;

    if ( stat(filename, &buf) != -1 ) {
        if ( S_ISDIR(buf.st_mode) )
            errno = EISDIR;
        else if ( ! S_ISREG(buf.st_mode) )
            errno = ENOTSUP;
        else if ( geteuid() == 0 &&
                ( buf.st_uid != 0 || buf.st_mode & ( S_IWGRP | S_IWOTH ) ) )
            /* When running as root, we will not allow reading a limits
             * file writable by anyone else than root itself. */
            errno = EPERM;
        else
            f = fopen(filename, "r");
    }

    return f;
}

#define LIMIT_LINE_LENGTH   128
#define USER_NAME_LENGTH    32

/**
 * Gets the resource limits for the specified user.
 * This function parses the limits(5) file and gets the resource limits
 * it specifies for the specified user. If that user cannot be found in
 * the file, and the file contains a default line, the default limits
 * are used.
 *
 * @param user The username to look for.
 * @param rlimits An array of rlimit_setting structures to be filled.
 * @param len Size of the \a rlimits array.
 * @return The number of resource limits, or -1 if an error occured.
 */
static ssize_t
get_limits_for_user(const char             *user,
                    struct rlimit_setting  *rlimits,
                    size_t                  len)
{
    FILE *f;
    char lbuf[LIMIT_LINE_LENGTH+1] = "", nbuf[USER_NAME_LENGTH+1];
    int user_found = 0, default_or_group_found = 0;
    ssize_t n = 0;

    if ( user == NULL || rlimits == NULL ) {
        errno = EINVAL;
        return -1;
    }

    if ( (f = open_limits_file(LIMITS_FILE)) == NULL )
        return -1;

    while ( ! feof(f) ) {
        if ( get_first_field(f, nbuf, sizeof(nbuf)) == -1 )
            continue;

        if ( strcmp(nbuf, "*") == 0 || user_in_group(user, nbuf) == 0 )
            default_or_group_found = get_line(f, lbuf, sizeof(lbuf)) != -1;
        else if ( strcmp(nbuf, user) == 0 )
            user_found = get_line(f, lbuf, sizeof(lbuf)) != -1;
        else
            discard_line(f);

#ifndef ENABLE_MERGED_LIMITS
        /*
         * The limits(5) man page states that only the FIRST line
         * matching a given username is to be considered, so we can
         * stop parsing as soon as we have found such a line.
         *
         * On the contrary, in the absence of a line matching the
         * username, the LAST default line or @group line takes
         * precedence, so we must continue parsing until EOF.
         */
        if ( user_found )
            break;
#endif
    }
    fclose(f);

    if ( user_found || default_or_group_found )
        if ( (n = parse_limit_string(lbuf, rlimits, len)) == -1 )
            errno = EBADMSG;

    if ( n >= 0 )
        rlimits[n].resource = -1;

    return n;
}

static int
set_limits(struct rlimit_setting *rlimits, pid_t pid)
{
    struct rlimit rlim;

#ifndef HAVE_PRLIMIT
    if ( pid != getpid() ) {
        errno = ENOSYS;
        return -1;
    }
#endif

    while ( rlimits->resource != -1 ) {
        switch ( rlimits->resource ) {
        case PLIMIT_UMASK:
            /* We cannot set the umask for another process
             * than our own. */
            if ( pid == getpid() )
                umask(rlimits->limit);
            break;

        case PLIMIT_LOGIN:
            /* Max number of logins; this is not a login program, so we
             * are not in a position to enforce that limit. */
            break;

        case PLIMIT_PRIORITY:
            /* setpriority expects a value between -20 and 19. */
            if ( setpriority(PRIO_PROCESS, pid, 20 - rlimits->limit) == -1 )
                return -1;
            break;

        default:
            rlim.rlim_cur = rlim.rlim_max = rlimits->limit;
#ifdef HAVE_PRLIMIT
            if ( prlimit(pid, rlimits->resource, &rlim, NULL) == -1 )
#else
            if ( setrlimit(rlimits->resource, &rlim) == -1 )
#endif
                return -1;
            break;
        }

        rlimits += 1;
    }

    return 0;
}

static void
print_ulimit_command(struct rlimit_setting *rlimits)
{
    printf("ulimit");

    while ( rlimits->resource != -1 ) {
        switch ( rlimits->resource ) {
#define ULIMIT(symbol, unused1, letter, unused2, factor)                \
        case symbol:                                                    \
            if ( letter != 0 ) {                                        \
                if ( rlimits->limit != RLIM_INFINITY )                  \
                    printf(" -%c %lu", letter, rlimits->limit / factor);\
                else                                                    \
                    printf(" -%c unlimited", letter);                   \
            }                                                           \
            break;
#include "ulimits.h"
#undef ULIMIT
        }
        rlimits += 1;
    }

    putchar('\n');
}


/* Helper functions. */

#define IS_SETUID_ROOT()    (getuid() != 0 && geteuid() == 0)

/**
 * Get the real user id of the given process.
 * This function spawns /bin/ps to obtain the real user id associated
 * with the specified process id.
 *
 * @param pid The process id to query.
 * @param ruid A pointer to a uid_t, to be filled with the real user id.
 * @return 0 if successful, or -1 if an error occured.
 */
static int
get_process_ruid(pid_t pid, uid_t *ruid)
{
    FILE *ps;
    char buffer[64];
    int ret = -1;

    if ( snprintf(buffer, 64, PS_PATH " -o ruid= %u", pid) >= 64 ) {
        errno = ERANGE;
        return -1;
    }

    if ( (ps = popen(buffer, "r")) != NULL ) {
        if ( fscanf(ps, "%u", ruid) == 1 )
            ret = 0;
        else
            errno = ESRCH;
        pclose(ps);
    }

    return ret;
}

/**
 * Check whether the user is allowed to alter the specified process.
 *
 * @param pid The process to query.
 * @return 0 if authorization is granted, -1 otherwise.
 */
static int
authorize(pid_t pid)
{
    uid_t puid;

    if ( IS_SETUID_ROOT() ) {
        /* Make sure the target process belongs to the real user. */

        if ( get_process_ruid(pid, &puid) == -1 )
            return -1;

        if ( puid != getuid() ) {
            errno = EPERM;
            return -1;
        }
    }

    return 0;
}

static unsigned
get_uinteger_or_die(const char *arg)
{
    unsigned val;
    char *endptr;

    errno = 0;
    val = strtoul(arg, &endptr, 10);
    if ( errno != 0 || endptr == arg )
        errx(EXIT_FAILURE, "invalid argument, integer expected: %s", arg);

    return val;
}


/* Main function. */

int
main(int argc, char **argv)
{
    struct rlimit_setting rlimits[16];
    struct passwd *pwd;
    pid_t pid;
    int n, prnt;

    struct option options[] = {
        { "help",    0, NULL, 'h' },
        { "version", 0, NULL, 'v' },
        { "ulimit",  0, NULL, 'c' },
        { "set",     0, NULL, 's' },
        { "process", 1, NULL, 'p' },
        { NULL,      0, NULL, 0 }
    };

    setprogname(argv[0]);
    pid = -1;
    prnt = 0;

    while ( (n = getopt_long(argc, argv, "hvcsp:", options, NULL)) != -1 ) {
        switch ( n ) {
        case 'h':
            usage(EXIT_SUCCESS);
            break;

        case '?':
            usage(EXIT_FAILURE);
            break;

        case 'v':
            info();
            break;

        case 'c':
            prnt = 1;
            break;

        case 's':
            prnt = 0;
            break;

        case 'p':
            pid = get_uinteger_or_die(optarg);
            break;
        }
    }

    if ( (pwd = getpwuid(getuid())) == NULL )
        err(EXIT_FAILURE, "cannot get user name");

    if ( get_limits_for_user(pwd->pw_name, rlimits, sizeof(rlimits)) == -1 )
        err(EXIT_FAILURE, "cannot get user limits");

    if ( prnt )
        print_ulimit_command(rlimits);
    else if ( argc - optind >= 1 ) {
        if ( set_limits(rlimits, getpid()) == -1 )
            err(EXIT_FAILURE, "cannot set limits");

        if ( IS_SETUID_ROOT() ) {
            /* Don't exec anything without droping privileges first. */
            if ( setgid(getgid()) == -1 || setuid(getuid()) == -1 )
                err(EXIT_FAILURE, "cannot drop privileges");

            if ( setuid(0) != -1 )
                err(EXIT_FAILURE, "cannot effectively drop privileges");
        }

        execvp(argv[optind], &argv[optind]);
        err(EXIT_FAILURE, "cannot execute program");
    }
    else {
        if ( pid == -1 )
            pid = getppid();
        else if ( authorize(pid) == -1 )
            err(EXIT_FAILURE, "cannot set limits for process %d", pid);

        if ( set_limits(rlimits, pid) == -1 )
            err(EXIT_FAILURE, "cannot set limits");
    }

    return EXIT_SUCCESS;
}
